/**
 * \file HkxPosControl.cpp
 * \brief Servo position control of the Arduino library for HerkuleX servo (DRS-0101 and DRS-0201 models)
 * \author Benoit Puel (contact@benoitpuel.com)
 * \version 1.1
 * \date 26 December 2015
 * \copyright This project is released under the GNU GPL V3.
 */

#include "HkxContRotation.h"


HkxContRotation::HkxContRotation(uint8_t ID, HkxCommunication& herkXCom, HkxPrint& print) : _herkXCom(herkXCom), _print(print), _className(F("HkxContRotation::")){
  if(ID > 0xFD){
    errorPrint(F("HkxContRotation > ID not correct"));
    return;
  }

  this->_id = ID;
  this->_statusED = HkxStatus();
  this->_torqueControl = HKX_TORQUE_FREE;
  this->_ledControl = HKX_LED_OFF;
  this->_connected = false;
 
  tryConnect();
}


boolean HkxContRotation::tryConnect(){
  uint8_t error = _herkXCom.statusRequest(this->_id, _statusED);
  if (error==2){
    String message = F("tryConnect > Servo ID=");
    message += this->_id;
    message += F(" is not connected");
    warningPrint(message);
    _connected = false;
    return false;
  } else {
    if(!_connected){ //if the servo was not connected before
      _connected = true;
      // Get setting data from the RAM
      uint8_t ramDataStart = 0;
      uint8_t ramDataLength = 2;
      uint8_t ramData[ramDataLength];
      uint8_t errorRAM = _herkXCom.readRequest(this->_id, HKX_RAM, ramDataStart, ramDataLength, ramData, _statusED);

      if(ramData[1] != 1){ // Force ACK Policy to be 1
        uint8_t ACKdata[] = {1};
        _herkXCom.writeRequest(this->_id, HKX_ROM, 7, 1, ACKdata);
        _herkXCom.writeRequest(this->_id, HKX_RAM, 1, 1, ACKdata);
      }

      // if the Servo has an error, then clean
      if (this->_statusED.isError(HKX_STAT_ALL) || this->_statusED.isDetail(HKX_STAT_ALL)){
        clearStatus();
      }
    }
    return true;
  }
}


uint8_t HkxContRotation::updateStatus(){
  if(!_connected){
    boolean connected = tryConnect();
    if(!connected){
      return 2;
    }
  }
  uint8_t error = _herkXCom.statusRequest(this->_id, _statusED);
  if(error == 2){// no data received
    String message = F("updateStatus > Servo ID=");
    message += this->_id;
    message += F(" is not connected");
    _connected = false;
    return 2;
  }
  if(error == 3){// data received not consistent
    return 3;
  }

  return 0;
}


uint8_t HkxContRotation::getStatus(HkxStatus& statusED, boolean update){
  if(!_connected){
    boolean connected = tryConnect();
    if(!connected){
      return 2;
    }
  }

  uint8_t error;
  if(update){
    error = updateStatus();
  }else{
    error = 0;
  }
  
  uint8_t statusRaw[2];
  this->_statusED.getRawData(statusRaw);
  statusED.setRawData(statusRaw);
  
  return error;
}


uint8_t HkxContRotation::clearStatus(){
  if(!_connected){
    boolean connected = tryConnect();
    if(!connected){
      return 2;
    }
  }
  // Set the status to 0
  uint8_t clearedStatus[] = {0x00 , 0x00};
  _herkXCom.writeRequest(this->_id, HKX_RAM, 48, 2, clearedStatus);
  // Switch back the torque and the LED to their previous status
  uint8_t torqueLEDControl[] = {this->_torqueControl, this->_ledControl};
  _herkXCom.writeRequest(this->_id, HKX_RAM, 52, 2, torqueLEDControl);
  
  return 0;
}

uint8_t HkxContRotation::setTorqueLEDControl(HkxMaybe<hkxTorqueControl> newTorqueControl, HkxMaybe<hkxLEDControl> newLEDControl){
  if(!_connected){
    boolean connected = tryConnect();
    if(!connected){
      return 2;
    }
  }

  if (newTorqueControl.isEmpty() && newLEDControl.isEmpty()){
    return 1;
  }

  uint8_t sizeData = 0;
  if (!newTorqueControl.isEmpty() && !newLEDControl.isEmpty()){
    sizeData = 2;
  } else {
    sizeData = 1;
  }

  uint8_t data[sizeData];
  uint8_t start;
  if (!newTorqueControl.isEmpty() && !newLEDControl.isEmpty()){
    _torqueControl = newTorqueControl.getValue();
    _ledControl = newLEDControl.getValue();
    data[0] = newTorqueControl.getValue();
    data[1] = newLEDControl.getValue();
    start = 52;
  } else {
    if (newTorqueControl.isEmpty()){
      data[0] = newLEDControl.getValue();
      start = 53;
      _ledControl = newLEDControl.getValue();
    } else {
      data[0] = newTorqueControl.getValue();
      start = 52;
    _torqueControl = newTorqueControl.getValue();
    }
  }

  _herkXCom.writeRequest(_id, HKX_RAM, start, sizeData, data);
    
  return 0;
}


uint8_t HkxContRotation::moveRotation(int16_t pwm, HkxMaybe<uint16_t> playTime, HkxMaybe<hkxLEDControl> LEDControl){
  if(!_connected){
    boolean connected = tryConnect();
    if(!connected){
      return 2;
    }
  }

  if(abs(pwm) > 1023){
    errorPrint(F("moveRotation > Input not correct"));
    return 1;
  }

  uint16_t pwmRaw[] = {pwm<0?((uint16_t)(-pwm))|0x4000:pwm}; // negative sign on the 14th bit

  // current torque and LED control
  hkxTorqueControl currentTorqueControl = this->_torqueControl;
  hkxLEDControl currentLEDControl = this->_ledControl;
  // new LED control
  hkxLEDControl newLEDControl = LEDControl.isEmpty()?_ledControl:LEDControl.getValue();
  setLEDControlVariable(newLEDControl);
  // torque on if not activated
  if (currentTorqueControl != HKX_TORQUE_ON){
    setTorqueLEDControl(HKX_TORQUE_ON, HKX_NO_VALUE);
  }

  // data for the sJog command
  uint8_t set[] = {2}; // mode continuous rotation
  set[0] |= newLEDControl<<2;
  uint8_t IDs[] = {this->_id};
  _ledControl = newLEDControl;
  _herkXCom.sJogRequest(0, 1, pwmRaw, set, IDs);// playtime is not considered by the servo for continuous rotation
  

  if(!playTime.isEmpty()){
    // wait stop
    delay(playTime.getValue());
    // stop the rotation
    _herkXCom.sJogRequest(0, 1, 0, set, IDs);
    // back to previous torque and LED states
    if (currentTorqueControl != HKX_TORQUE_ON || currentLEDControl != newLEDControl){
      this->setTorqueLEDControl(currentTorqueControl, currentLEDControl);
    }
  }

  return 0;
}


uint8_t HkxContRotation::getBehaviour(HkxMaybe<uint16_t> inputVoltage, HkxMaybe<uint16_t> temperature, HkxMaybe<int16_t> velocity, HkxMaybe<int16_t> PWM){
  if(!_connected){
    boolean connected = tryConnect();
    if(!connected){
      return 2;
    }
  }
  uint8_t behaviousData[20];
  uint8_t error = _herkXCom.readRequest(_id, HKX_RAM, 54, 20, behaviousData, _statusED);

  if(error == 2){// no data received
    String message = F("getBehaviour > Servo ID=");
    message += this->_id;
    message += F(" is not connected");
    warningPrint(message);
    this->_connected = false;
    return 2;
  }
  if(error == 3){// data received not consistent
    return 3;
  }
  
  if(!inputVoltage.isEmpty()){
    inputVoltage.setValue(HkxUnitConversion::voltageRawToValue(behaviousData[0]));
  }
  if(!temperature.isEmpty()){
    temperature.setValue(HkxUnitConversion::temperatureRawToValue(behaviousData[1]));
  }
  if(!velocity.isEmpty()){
    velocity.setValue(HkxUnitConversion::velocityRawToValue(behaviousData[9]<<8 | behaviousData[8]));
  }
  if(!PWM.isEmpty()){
    PWM.setValue(behaviousData[11]<<8 | behaviousData[10]);
  }

  return 0;
}


uint8_t HkxContRotation::reboot(){
  if(!_connected){
    boolean connected = tryConnect();
    if(!connected){
      return 2;
    }
  }
  _herkXCom.rebootRequest(this->_id);
  return 0;
}


