/**
 * \file HkxGroupContRotation.cpp
 * \brief Group of servos for continuous rotation of the Arduino library for HerkuleX servo (DRS-0101 and DRS-0201 models)
 * \author Benoit Puel (contact@benoitpuel.com)
 * \version 1.1
 * \date 26 December 2015
 * \copyright This project is released under the GNU GPL V3.
 */

#include "HkxGroupContRotation.h"


HkxGroupContRotation::HkxGroupContRotation(uint8_t length, HkxContRotation* arrayServos[], HkxPrint& print) : _herkXCom(arrayServos[0]->getCommunication()), _print(print), _length(length), _className(F("HkxGroupContRotation::")){
  this->_servos = new HkxContRotation*[length];
  uint8_t ids[length];
  for(int i = 0; i < length ; i++){
    this->_servos[i] = arrayServos[i];
    ids[i] = this->_servos[i]->getID();
    
    // Check the communication is the same for all servos
    if(i > 0){
      if(&(this->_servos[0]->getCommunication()) != &_herkXCom){
        errorPrint(F("HkxGroupContRotation > All servos need the same communication"));
        return;
      }
    }
    
    // Check IDs are different for the different servos
    for(int j = 0 ; j < i-1 ; j++){
      if(ids[i] == ids[j]){
        String messageID = F("HkxGroupContRotation > Servo id=");
        messageID += ids[i];
        messageID += F(" is set twice in the group");
        warningPrint(messageID);
      }
    }
    String messageADD = F("HkxGroupContRotation > Servo id=");
    messageADD += ids[i];
    messageADD += F(" is added to the group");
    infoPrint(messageADD);
  }
  if(length > 53){
    warningPrint(F("HkxGroupContRotation > Too many servos for asynchrone and synchrone moves"));
    return;
  } else if (length > 43){
    errorPrint(F("HkxGroupContRotation > Too many servos for asynchrone moves"));
    return;
  }
}

uint8_t HkxGroupContRotation::setAllTorqueLEDControl(HkxMaybe<hkxTorqueControl> newTorqueControl, HkxMaybe<hkxLEDControl> newLEDControl){
  uint8_t error;
  boolean failConnect;
  uint8_t nbTryConnect;
  for(int i = 0 ; i < this->_length ; i++){
    failConnect = false;
    nbTryConnect = 0;
    do{
      nbTryConnect++;
      error = this->_servos[i]->setTorqueLEDControl(newTorqueControl, newLEDControl);
      switch(error){
      case 0: //OK
        failConnect = false;
        break;
      case 1: //Input not correct
        return error;
        break;
      case 2: //problem of servo connection
        String message = F("setAllTorqueLEDControl > Servo id=");
        message += this->_servos[i]->getID();
        message += F(" failed to connect (try ");
        message += nbTryConnect;
        message += F(")");
        warningPrint(message);
        failConnect = true;
        if(nbTryConnect == HKX_NUMBER_CONNECT_TRIALS){ // HKX_NUMBER_CONNECT_TRIALS defined in HkxCommunication.h
          return error;
        }
        break;
      }
    } while(failConnect);
  }
  return 0;
}


uint8_t HkxGroupContRotation::moveAllRotation(int16_t pwm[], HkxMaybe<uint16_t> playTime, HkxMaybe<hkxLEDControl> LEDControl[]){
  if(_length > 53){
    errorPrint(F("moveAllRotation > Too many servos (> 53)"));
    return 4;
  }

  // current torque and LED control
  hkxTorqueControl currentTorqueControl[this->_length];
  hkxLEDControl currentLEDControl[this->_length];
  hkxLEDControl newLEDControl[this->_length];
  // torque on if not activated
  for(int i=0 ; i < this->_length ; i++){
    currentTorqueControl[i] = this->_servos[i]->getTorqueControl();
    currentLEDControl[i] = this->_servos[i]->getLEDControl();
    newLEDControl[i] = LEDControl[i].isEmpty()?currentLEDControl[i]:LEDControl[i].getValue();
    if (currentTorqueControl[i] != HKX_TORQUE_ON){
      this->_servos[i]->setTorqueLEDControl(HKX_TORQUE_ON, HKX_NO_VALUE);
    }
  }

  uint8_t nbTryConnect;
  uint8_t error;
  // data for the sJog command
  uint16_t pwmRaw[this->_length];
  uint8_t set[this->_length];
  uint8_t IDs[this->_length];
  for(int i=0 ; i < this->_length ; i++){
    if(!this->_servos[i]->isConnected()){
      nbTryConnect = 0;
      boolean connected;
      do{
        nbTryConnect++;
        String message = F("moveAllRotation > Servo id=");
        message += this->_servos[i]->getID();
        message += F(" failed to connect (try ");
        message += nbTryConnect;
        message += F(")");
        warningPrint(message);
        connected = this->_servos[i]->tryConnect();
      } while(!connected && nbTryConnect < HKX_NUMBER_CONNECT_TRIALS); // HKX_NUMBER_CONNECT_TRIALS defined in HkxCommunication.h     
      if(nbTryConnect == HKX_NUMBER_CONNECT_TRIALS){
        return 2;
      }
    }
    pwmRaw[i] = pwm[i]<0?((uint16_t)(-pwm[i]))|0x4000:pwm[i]; // negative sign on the 14th bit

    if(abs(pwm[i]) > 1023){
      String message = F("moveAllRotation > Try to move the servo id=");
      message += this->_servos[i]->getID();
      message += F(" with a PWM out of the range");
      errorPrint(message);
      return 1;
    }
    set[i] = 2;
    set[i] |= newLEDControl[i]<<2;
    IDs[i] = this->_servos[i]->getID();
    this->_servos[i]->setLEDControlVariable(newLEDControl[i]);
  } 

  _herkXCom.sJogRequest(0, this->_length, pwmRaw, set, IDs);// playtime is not considered by the servo for continuous rotation

  if(!playTime.isEmpty()){
    delay(playTime.getValue());
    // stop the rotation
    uint16_t pwmNull[this->_length];
    memset(pwmNull,0,this->_length); // set all values to 0
    _herkXCom.sJogRequest(0, this->_length, pwmNull, set, IDs);  
    // back to previous torque and LED states
    for(int i=0 ; i < this->_length ; i++){
      if (currentTorqueControl[i] != HKX_TORQUE_ON || currentLEDControl[i] != newLEDControl[i]){ 
        this->_servos[i]->setTorqueLEDControl(currentTorqueControl[i], currentLEDControl[i]);
      }
    }
  }

  return 0;
}


